/*
 * Copyright (c) 1991-2023 by STEP Tools Inc. 
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and
 * its documentation is hereby granted, provided that this copyright
 * notice and license appear on all copies of the software.
 * 
 * STEP TOOLS MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
 * SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT. STEP TOOLS
 * SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A
 * RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES.
 *
 * Author: David Loffredo (loffredo@steptools.com)
 */
#include <stp_schema.h>
#include <rosemath.h>
#include "make_polyshell.h"

// CREATE A SWEPT POLYGON SHELL WITH HOLES - pass in a list of 2D
// points and a thickness.  Use those points to define a polygonal
// face in the XY plane, define another one offset by thickness in the
// -Z direction, then sew everything up as a closed brep shell.
//
// Pass back the shell.  Pass back the top and bottom face in case we
// want to define any holes in them.
//

ROSE_DECLARE_VECTOR(EdgeCurveList,stp_edge_curve);


stp_vertex_point * make_vertex(
    stp_cartesian_point * pt
    )
{
    stp_vertex_point * v = pnewIn(pt->design()) stp_vertex_point;
    v->name("");
    v->vertex_geometry (pt);
    return v;
}


static stp_cartesian_point * get_point ( stp_vertex * v )
{
    // need to cast down because everything is generalized,
    // but we know that we are dealing with the more specific
    // types.
    stp_vertex_point * vpt = ROSE_CAST(stp_vertex_point,v);
    return ROSE_CAST(stp_cartesian_point,vpt->vertex_geometry());
}

static void get_position ( double pos[3], stp_cartesian_point * pt )
{
    ListOfDouble * lst=pt-> coordinates();
    pos[0]=lst->get(0);
    pos[1]=lst->get(1);
    pos[2]=lst->get(2);
}
static void get_position ( double pos[3], stp_vertex * v )
{
    get_position(pos, get_point(v));
}

stp_cartesian_point * make_point(
    RoseDesign * d,
    double x, double y, double z
    )
{
    stp_cartesian_point * pt = pnewIn(d) stp_cartesian_point;
    pt->name("");
    pt->coordinates()->add (x);	
    pt->coordinates()->add (y);	
    pt->coordinates()->add (z);
    return pt;
}

stp_cartesian_point * make_bottom_point(
    stp_cartesian_point * pt,
    double thickness
    )
{
    // if we wanted faces that are tilted, we could pass in a
    // direction and do a bit of projection.  As it is, the bottom
    // point is just shifted down in the Z direction.
    double xyz[3];
    get_position(xyz, pt);
    xyz[2] -= thickness;
    return make_point(pt->design(), xyz[0], xyz[1], xyz[2]);
}



    
stp_axis2_placement_3d * make_ap3d(
    stp_cartesian_point * pt,
    double zdir[3],
    double xdir[3]
    )
{
    RoseDesign * d = pt->design();
    
    stp_direction * zaxis = pnewIn(d) stp_direction;
    zaxis->name("");
    zaxis->direction_ratios()->add(zdir[0]);
    zaxis->direction_ratios()->add(zdir[1]);
    zaxis->direction_ratios()->add(zdir[2]);

    stp_direction * xaxis = pnewIn(d) stp_direction;
    xaxis->name("");
    xaxis->direction_ratios()->add(xdir[0]);
    xaxis->direction_ratios()->add(xdir[1]);
    xaxis->direction_ratios()->add(xdir[2]);
    
    stp_axis2_placement_3d * ap3d = pnewIn(d) stp_axis2_placement_3d;
    ap3d->name("");
    ap3d->location (pt);
    ap3d->axis (zaxis);
    ap3d->ref_direction (xaxis);
    return ap3d;
}


stp_edge_curve * make_edge(
    stp_vertex * vstart,
    stp_vertex * vfinish
    ) 
{
    // get the points and do some basic math to find out the distance
    // and direction so that we can make a line, then wrap in an edge
    // curve.
    RoseDesign * d = vstart->design();

    double xyzstart[3];
    double xyzfinish[3];

    get_position(xyzstart, vstart);
    get_position(xyzfinish, vfinish);
	    
    // compute the direction vector and normalize
    double calcdir[3];
    rose_vec_diff(calcdir, xyzfinish, xyzstart);
    rose_vec_normalize(calcdir, calcdir);

    stp_direction * td1 = pnewIn(d) stp_direction;
    td1->name("");
    td1->direction_ratios()->add(calcdir[0])->add(calcdir[1])->add(calcdir[2]);

    stp_vector * tve1 = pnewIn(d) stp_vector;
    tve1->name("");
    tve1->orientation (td1);
    tve1->magnitude (rose_pt_distance(xyzstart, xyzfinish));

    stp_line * tln1 = pnewIn(d) stp_line;
    tln1->name("");
    tln1->pnt (get_point(vstart));
    tln1->dir (tve1);

    stp_edge_curve * ec = pnewIn(d) stp_edge_curve;
    ec->name("");
    ec->edge_geometry (tln1);
    ec->same_sense (1);
    ec->edge_start (vstart);
    ec->edge_end (vfinish);

    return ec;
}

stp_edge_curve * make_edge(
    stp_vertex * vstart,
    stp_vertex * vfinish,
    stp_circle * geom
    ) 
{
    RoseDesign * d = vstart->design();

    stp_edge_curve * ec = pnewIn(d) stp_edge_curve;
    ec->name("");
    ec->edge_geometry (geom);
    ec->same_sense (1);
    ec->edge_start (vstart);
    ec->edge_end (vfinish);

    return ec;
}


stp_oriented_edge * make_oriented(
    stp_edge_curve * ec,
    int sense
    )
{
    RoseDesign * d = ec->design();

    stp_oriented_edge * oe = pnewIn(d) stp_oriented_edge;
    oe-> name("");
    oe-> edge_element (ec);
    oe-> orientation (sense);
    return oe;
}




// make a face with a planar surface, since the loop points are
// planar, compute the plane axis placement from that.
static stp_advanced_face * make_planar_face(
    stp_edge_loop * lp
    )
{
    RoseDesign * d = lp->design();
    
    // get three points for computing the plane normal.  Watch for
    // reversed orientation on the edges.
    stp_cartesian_point * p1;
    stp_cartesian_point * p2;
    stp_cartesian_point * p3;
    stp_oriented_edge * oe = lp->edge_list()->get(0);

    if (oe->orientation()) {
	p1 = get_point(oe->edge_element()->edge_start());
	p2 = get_point(oe->edge_element()->edge_end());
    }
    else {
	p2 = get_point(oe->edge_element()->edge_start());
	p1 = get_point(oe->edge_element()->edge_end());
    }

    oe = lp->edge_list()->get(1);
    if (oe->orientation()) {
	// start should share p2
	p3 = get_point(oe->edge_element()->edge_end());
    }
    else {
	p3 = get_point(oe->edge_element()->edge_start());
	// end should share p2
    }

    // compute direction of face.  P1,P2,P3 should be counter
    // clockwise and so P2->P3 x P2->P1 gives the face normal.
    double p1coords[3];
    double p2coords[3];
    double p3coords[3];

    get_position(p1coords, p1);
    get_position(p2coords, p2);
    get_position(p3coords, p3);

    double vecp2p3[3];
    double vecp2p1[3];

    rose_vec_diff(vecp2p3, p3coords, p2coords);
    rose_vec_diff(vecp2p1, p1coords, p2coords);
    
    double facenorm[3];

    rose_vec_cross(facenorm, vecp2p3, vecp2p1);
    rose_vec_normalize(facenorm, facenorm);
    rose_vec_normalize(vecp2p3, vecp2p3);  // use as xaxis for placement
    
    stp_plane * afp = pnewIn(d) stp_plane;
    afp->name("");
    afp->position (make_ap3d(p2, facenorm, vecp2p3));

    // grab three points and make a direction and AP

    stp_face_outer_bound *bnd = pnewIn(d) stp_face_outer_bound;
    bnd->name("");
    bnd->bound (lp);
    bnd->orientation (1);

    stp_advanced_face * af = pnewIn(d) stp_advanced_face;
    af->name("");
    af->face_geometry (afp);
    af->same_sense (1);
    af->bounds()->add (bnd);
    return af;
}


void make_polygon_shell (
    stp_closed_shell ** ret_shell, 
    stp_advanced_face ** ret_top, 
    stp_advanced_face ** ret_bottom, 
    RoseDesign * d,
    ListOfstp_cartesian_point * xypts, 
    double thickness
    )
{
    unsigned i = 0;
    stp_cartesian_point * pttop = xypts->get(i++);
    stp_cartesian_point * ptbot = make_bottom_point (pttop, thickness);
    
    stp_vertex_point * vtop1 = make_vertex(pttop);
    stp_vertex_point * vbot1 = make_vertex(ptbot);
    stp_vertex_point * vtop2, *vbot2;

    EdgeCurveList top_edges;
    EdgeCurveList bot_edges;

    // Walk the list of points creating edges and a side face for each
    // pair and sewing them together into the face loops.
    while ( i<xypts->size() )
    {
	pttop = xypts->get(i++);
	ptbot = make_bottom_point (pttop, thickness);

	vtop2 = make_vertex(pttop);
	vbot2 = make_vertex(ptbot);

	top_edges.append (make_edge(vtop1, vtop2));
	bot_edges.append (make_edge(vbot1, vbot2));

	vtop1 = vtop2;
	vbot1 = vbot2;
    }

    // close the loop
    top_edges.append (
	make_edge(top_edges[top_edges.size()-1]->edge_end(),
		  top_edges[0]->edge_start())
	);

    bot_edges.append (
	make_edge(bot_edges[bot_edges.size()-1]->edge_end(),
		  bot_edges[0]->edge_start())
	);


    // All of the edges for the top and bottom perimeter are now
    // present, build faces from them and assemble into a shell
    
    stp_closed_shell *shell = pnewIn(d) stp_closed_shell;
    shell->name("");

    // Go around the loop and add edges to connect the top and bottom,
    // and make advanced faces for everything.
    stp_edge_curve * first_edge = 0;
    stp_edge_curve * last_edge = 0;
    stp_edge_loop * top_loop = pnewIn(d) stp_edge_loop;
    stp_edge_loop * bot_loop = pnewIn(d) stp_edge_loop;

    top_loop->name("");
    bot_loop->name("");

    for (i=0; i<top_edges.size(); i++)
    {
	// Always make the side edges point top to bottom in the the
	// sweep direction.  This makes it easier to get the direction
	// flipping correct when we loop back around.
	// 1-->2 	top
	// 3-->4	bottom
	stp_edge_curve * etop = top_edges[i];
	stp_edge_curve * ebot = bot_edges[i];

	stp_vertex * v1 = etop->edge_start();
	stp_vertex * v2 = etop->edge_end();
	stp_vertex * v3 = ebot->edge_start();
	stp_vertex * v4 = ebot->edge_end();

	stp_edge_curve * e9oclock = last_edge? last_edge: make_edge(v1, v3);
	stp_edge_curve * e3oclock =
	    (i < (top_edges.size()-1))? make_edge(v2, v4): first_edge;

	// save for wrapping
	if (!first_edge) first_edge = e9oclock;
	last_edge = e3oclock;

	// make an edge loop, fill with the edges, then embed in an
	// advanced face.  first and last one flipped
	stp_edge_loop * floop = pnewIn(d) stp_edge_loop;
	floop->name("");
	floop->edge_list()->add (make_oriented(etop, 0));
	floop->edge_list()->add (make_oriented(e9oclock, 1));
	floop->edge_list()->add (make_oriented(ebot, 1));
	floop->edge_list()->add (make_oriented(e3oclock, 0));
	// make face add to face set
	shell->cfs_faces()->add (make_planar_face(floop));

	// the edges for the top and bottom faces.  The edges for the
	// bottom are reversed because it points the other way, so
	// flip the edge and insert to the list in reverse order.
	top_loop->edge_list()->add (make_oriented(top_edges[i], 1));
	bot_loop->edge_list()->insert (make_oriented(bot_edges[i], 0), 0);

    }
    (*ret_shell) = shell;
    (*ret_top) = make_planar_face(top_loop);
    (*ret_bottom) = make_planar_face(bot_loop);

    shell->cfs_faces()->add (*ret_top);
    shell->cfs_faces()->add (*ret_bottom);
}



void make_circular_hole (
    stp_closed_shell * shell, 
    stp_advanced_face * top_face, 
    stp_advanced_face * bottom_face, 
    double x, double y,
    double rad,
    double thickness
    )
{
    RoseDesign * d = top_face->design();

    double updir[3] = {0, 0, 1};
    double downdir[3] = {0, 0, -1};
    double xdir[3] = {1, 0, 0};
    
    // Holes in step are commonly split into two half cylinders.  Not
    // the only way to do it, but the most common.

    stp_cartesian_point * ptctrtop = make_point (d, x, y);
    stp_cartesian_point * ptltop = make_point (d, x-rad, y);
    stp_cartesian_point * ptrtop = make_point (d, x+rad, y);

    stp_cartesian_point * ptctrbot = make_bottom_point (ptctrtop, thickness);
    stp_cartesian_point * ptlbot = make_bottom_point (ptltop, thickness);
    stp_cartesian_point * ptrbot = make_bottom_point (ptrtop, thickness);
    
    stp_circle * ctop = pnewIn(d) stp_circle;
    ctop->name ("");
    ctop->radius (rad);
    ctop->position (pnewIn(d) stp_axis2_placement);
    ctop->position()->_axis2_placement_3d ( make_ap3d(ptctrtop, updir, xdir) );

    stp_circle * cbot = pnewIn(d) stp_circle;
    cbot->name ("");
    cbot->radius (rad);
    cbot->position (pnewIn(d) stp_axis2_placement);
    cbot->position()->_axis2_placement_3d ( make_ap3d(ptctrbot, downdir, xdir) );

    stp_cylindrical_surface * cyl = pnewIn(d) stp_cylindrical_surface;
    cyl-> name("");
    cyl-> radius(rad);
    cyl-> position (ctop->position()->_axis2_placement_3d());

    // Start makin faces
    stp_vertex_point * vltop = make_vertex(ptltop);
    stp_vertex_point * vrtop = make_vertex(ptrtop);
    stp_vertex_point * vlbot = make_vertex(ptlbot);
    stp_vertex_point * vrbot = make_vertex(ptrbot);

    stp_edge_curve * ecs[6];
    ecs[0] = make_edge(vltop, vrtop, ctop); 
    ecs[1] = make_edge(vrtop, vltop, ctop); 
    ecs[2] = make_edge(vlbot, vrbot, cbot); 
    ecs[3] = make_edge(vrbot, vlbot, cbot); 
    ecs[4] = make_edge(vltop, vlbot);  // the sides
    ecs[5] = make_edge(vrtop, vrbot); 

    // make an edge loop, fill with the edges, then embed in an
    // advanced face.  first and last one flipped
    stp_edge_loop * floop = pnewIn(d) stp_edge_loop;
    floop->name("");
    floop->edge_list()->add (make_oriented(ecs[0], 1));
    floop->edge_list()->add (make_oriented(ecs[5], 1));
    floop->edge_list()->add (make_oriented(ecs[3], 1));
    floop->edge_list()->add (make_oriented(ecs[4], 0));
    
    stp_face_outer_bound * fb = pnewIn(d) stp_face_outer_bound;
    fb->name("");
    fb->bound (floop);
    fb->orientation (1);

    stp_advanced_face * af = pnewIn(d) stp_advanced_face;
    af->name("");
    af->face_geometry (cyl);
    af->same_sense (0);
    af->bounds()->add (fb);

    shell->cfs_faces()->add (af);

    // make second face
    floop = pnewIn(d) stp_edge_loop;
    floop->name("");
    floop->edge_list()->add (make_oriented(ecs[1], 1));
    floop->edge_list()->add (make_oriented(ecs[4], 1));
    floop->edge_list()->add (make_oriented(ecs[2], 1));
    floop->edge_list()->add (make_oriented(ecs[5], 0));
    
    fb = pnewIn(d) stp_face_outer_bound;
    fb->name("");
    fb->bound (floop);
    fb->orientation (1);

    af = pnewIn(d) stp_advanced_face;
    af->name("");
    af->face_geometry (cyl);
    af->same_sense (0);
    af->bounds()->add (fb);

    shell->cfs_faces()->add (af);

    // now add top and bottom loops
    stp_edge_loop * top_loop = pnewIn(d) stp_edge_loop;
    stp_edge_loop * bot_loop = pnewIn(d) stp_edge_loop;

    top_loop->name("");
    top_loop->edge_list()->add (make_oriented(ecs[0], 0));
    top_loop->edge_list()->add (make_oriented(ecs[1], 0));

    bot_loop->name("");
    bot_loop->edge_list()->add (make_oriented(ecs[2], 0));
    bot_loop->edge_list()->add (make_oriented(ecs[3], 0));

    stp_face_bound * bnd = pnewIn(d) stp_face_bound;
    bnd->name("");
    bnd->bound (top_loop);
    bnd->orientation (1);
    top_face->bounds()->add (bnd);

    bnd = pnewIn(d) stp_face_bound;
    bnd->name("");
    bnd->bound (bot_loop);
    bnd->orientation (1);
    bottom_face->bounds()->add (bnd);
}




void make_polygon_hole (
    stp_closed_shell * shell, 
    stp_advanced_face * top_face, 
    stp_advanced_face * bottom_face, 
    ListOfstp_cartesian_point * xypts,
    double thickness
    )
{
    // This a copy of the make polygon shell with several exceptions.
    // First, we do not make top and bottom faces with an outer bound,
    // instead we make a plain bound and add it to the existing
    // top/bottom.  Second, the sense of the side faces is flipped
    // because we are doing a hole.
    
    RoseDesign * d = top_face->design();
    unsigned i = 0;
    stp_cartesian_point * pttop = xypts->get(i++);
    stp_cartesian_point * ptbot = make_bottom_point (pttop, thickness);
    
    stp_vertex_point * vtop1 = make_vertex(pttop);
    stp_vertex_point * vbot1 = make_vertex(ptbot);
    stp_vertex_point * vtop2, *vbot2;

    EdgeCurveList top_edges;
    EdgeCurveList bot_edges;

    // Walk the list of points creating edges and a side face for each
    // pair and sewing them together into the face loops.
    while ( i<xypts->size() )
    {
	pttop = xypts->get(i++);
	ptbot = make_bottom_point (pttop, thickness);

	vtop2 = make_vertex(pttop);
	vbot2 = make_vertex(ptbot);

	top_edges.append (make_edge(vtop1, vtop2));
	bot_edges.append (make_edge(vbot1, vbot2));

	vtop1 = vtop2;
	vbot1 = vbot2;
    }

    // close the loop
    top_edges.append (
	make_edge(top_edges[top_edges.size()-1]->edge_end(),
		  top_edges[0]->edge_start())
	);

    bot_edges.append (
	make_edge(bot_edges[bot_edges.size()-1]->edge_end(),
		  bot_edges[0]->edge_start())
	);


    // Go around the loop and add edges to connect the top and bottom,
    // and make advanced faces for everything.
    stp_edge_curve * first_edge = 0;
    stp_edge_curve * last_edge = 0;
    stp_edge_loop * top_loop = pnewIn(d) stp_edge_loop;
    stp_edge_loop * bot_loop = pnewIn(d) stp_edge_loop;

    top_loop->name("");
    bot_loop->name("");

    for (i=0; i<top_edges.size(); i++)
    {
	// Always make the side edges point top to bottom in the the
	// sweep direction.  This makes it easier to get the direction
	// flipping correct when we loop back around.
	// 1-->2 	top
	// 3-->4	bottom
	
	stp_edge_curve * etop = top_edges[i];
	stp_edge_curve * ebot = bot_edges[i];

	stp_vertex * v1 = etop->edge_start();
	stp_vertex * v2 = etop->edge_end();
	stp_vertex * v3 = ebot->edge_start();
	stp_vertex * v4 = ebot->edge_end();

	stp_edge_curve * e9oclock = last_edge? last_edge: make_edge(v1, v3);
	stp_edge_curve * e3oclock =
	    (i < (top_edges.size()-1))? make_edge(v2, v4): first_edge;

	// save for wrapping
	if (!first_edge) first_edge = e9oclock;
	last_edge = e3oclock;

	// make an edge loop, fill with the edges, then embed in an
	// advanced face.  first and last one flipped
	stp_edge_loop * floop = pnewIn(d) stp_edge_loop;
	floop->name("");
	floop->edge_list()->add (make_oriented(etop, 0));
	floop->edge_list()->add (make_oriented(e9oclock, 1));
	floop->edge_list()->add (make_oriented(ebot, 1));
	floop->edge_list()->add (make_oriented(e3oclock, 0));

	stp_advanced_face * af = make_planar_face(floop);
	af-> same_sense(0);  // flip because it is a hole

	shell->cfs_faces()->add (af);

	// the edges for the top and bottom faces.  The edges for the
	// bottom are reversed because it points the other way, so
	// flip the edge and insert to the list in reverse order.
	top_loop->edge_list()->add (make_oriented(top_edges[i], 1));
	bot_loop->edge_list()->insert (make_oriented(bot_edges[i], 0), 0);
    }

    // top and bottom faces exist, add this as another loop
    stp_face_bound * bnd = pnewIn(d) stp_face_bound;
    bnd->name("");
    bnd->bound (top_loop);
    bnd->orientation (1);
    top_face->bounds()->add (bnd);

    bnd = pnewIn(d) stp_face_bound;
    bnd->name("");
    bnd->bound (bot_loop);
    bnd->orientation (1);
    bottom_face->bounds()->add (bnd);
}
    

